
{*******************************************************}
{                                                       }
{       RichView                                        }
{       Adding table of contents (RTF version)          }
{                                                       }
{       Copyright (c) Sergey Tkachenko                  }
{       svt@trichview.com                               }
{       http://www.trichview.com                        }
{                                                       }
{*******************************************************}


unit RVTOCRTF;

interface

uses SysUtils, Classes, Graphics, RichView, PtblRV, RVClasses, CRVData,
     RVMarker, RVLabelItem, RVTable, RVStyle, RVScroll, RVReport, RVTypes;


{

  This procedure adds a table of contents (TOC) to rv.
  This TOC is designed for export in RTF.

  TOC is inserted in the position specified in ItemNo.
  Possible values of ItemNo:
  * 0 - inserting at the beginning of rv;
  * rv.ItemCount - inserting at the end of rv
  * any value from 1 to rv.ItemCount-1 - inserting before the item with this index;
    this item must have a page break (i.e. rv.PageBreaksBeforeItems[ItemNo] must
    be True), otherwise this function does nothing and returns False.

  This function returns True on successful inserting. If the document does not
  have headings with levels in range 1..Depth, this function does nothing,
  but still returns True.

  RVPrint must be formatted before the call of this function. After the call,
  it is not formatted (call RVPrint.FormatPages again before printing).

  Rv is not formatted after the call of this function (call rv.Format before
  displaying it).

  Title is added before the TOC, using TitleStyleNo and TitleParaNo styles.
  Depth is a maximal heading level for TOC.
  TextStyleNos and ParaStyleNos are arrays containing styles for TOC. They
  must have Depth items.
  TitleStyleNo[0] and TitleParaNo[0] are used for adding heading level 1 in TOC,
  TitleStyleNo[1] and TitleParaNo[2] are used for adding heading level 2 in TOC,
  and so on.

  TOC is added not as an editing operation! Undo is not possible. If called for
  TRichViewEdit, call ClearUndo method.

  // TitleStyleNo, TextStyleNos[] are indices in rv.Style.TextStyles collection.
  // TitleParaNo, ParaStyleNos[] are indices in rv.Style.ParaStyles collection.
}

function AddTableOfContents(rv: TCustomRichView; RVPrint: TRVPrint;
  Depth: Integer; const TextStyleNos, ParaStyleNos: array of Integer;
  ItemNo: Integer; const Title: TRVUnicodeString;
  TitleStyleNo, TitleParaNo: Integer): Boolean; overload;

{
  The same procedure for TRVReportHelper.
  Additional parameters:
  Canvas - a canvas that was used for RVReportHelper.Init(), or another canvas
    with the same resolution; can be nil if ItemNo=RVReportHelper.RichView.ItemCount
  PageWidth - width of pages (used in Init)
  FirstPageHeight - height of the first page (used in FormatNextPage)
  PageHeight - height of other pages (used in FormatNextPage)

  RVReportHelper must be formatted before the call of this function.
  After the call, it is not formatted.

}


function AddTableOfContents(RVReportHelper: TRVReportHelper; Canvas: TCanvas;
  PageWidth, FirstPageHeight, PageHeight: Integer;
  Depth: Integer; const TextStyleNos, ParaStyleNos: array of Integer;
  ItemNo: Integer; const Title: TRVUnicodeString;
  TitleStyleNo, TitleParaNo: Integer): Boolean; overload;

{
  This procedure searches for the first occurence of the paragraph
  with OutlineLevel=Level, and returns:
  - the style of this paragraph (in ParaNo),
  - the style of text in this paragraph (in StyleNo).
  Paragraphs without text are ignored.
  If there is no such paragraph, this procedure returns 0, 0.
  These values can be used as TitleStyleNo, TitleParaNo parameters for
  AddTableOfContents.
}

procedure GetStylesForOutlineLevel(rv: TCustomRichView; Level: Integer;
  var StyleNo, ParaNo: Integer);

implementation

uses
  RVStrFuncs;

type
  TTOCItem = class
    Text: TRVUnicodeString;
    Level, PageNo, ID: Integer;
  end;

function GetParagraphText(RVData: TCustomRVData; ItemNo: Integer): TRVUnicodeString;
  {..........................................}
  function GetItemText: String;
  begin
    if RVData.GetItemStyle(ItemNo)>=0 then
      Result := RVData.GetItemTextW(ItemNo)
    else if RVData.GetItemStyle(ItemNo)=rvsListMarker then
      Result := TRVMarkerItemInfo(RVData.GetItem(ItemNo)).DisplayString+' '
    else if RVData.GetItemStyle(ItemNo) = rvsTab then
      Result := ' '
    else if RVData.GetItem(ItemNo) is TRVLabelItemInfo then
      Result := TRVLabelItemInfo(RVData.GetItem(ItemNo)).Text
    else
      Result := '';
  end;
  {..........................................}
begin
  Result := GetItemText;
  inc(ItemNo);
  while (ItemNo<RVData.ItemCount) and not RVData.IsParaStart(ItemNo) do begin
    Result := Result+GetItemText;
    inc(ItemNo);
  end;
end;
{------------------------------------------------------------------------------}
function BuildTOCStructure(rv: TCustomRichView; RVPrint: TCustomRVPrint;
  Depth: Integer): TRVList;
  {.........................................................}
  procedure DoBuildTOC(RVData: TCustomRVData; TOCList: TRVList);
  var i, r, c, Level: Integer;
      table: TRVTableItemInfo;
      TOCItem: TTOCItem;
  begin
    for i := 0 to RVData.ItemCount-1 do begin
      if RVData.IsParaStart(i) then begin
        Level := rv.Style.ParaStyles[RVData.GetItemPara(i)].OutlineLevel;
        if (Level>0) and (Level<=Depth) then begin
          RVData.SetCheckpointInfo(i, RVEMPTYTAG,
            '_toc'+RVIntToStrW(TOCList.Count), False);
          TOCItem := TTOCItem.Create;
          TOCItem.Text := GetParagraphText(RVData, i);
          TOCItem.Level := Level-1;
          TOCItem.PageNo := RVPrint.GetPageNo(RVData, i, RVData.GetOffsBeforeItem(i));
          TOCItem.ID := TOCList.Count;
          TOCList.Add(TOCItem);
        end;
      end;
      if RVData.GetItemStyle(i)=rvsTable then begin
        table := TRVTableItemInfo(RVData.GetItem(i));
        for r := 0 to table.RowCount-1 do
          for c := 0 to table.ColCount-1 do
            if table.Cells[r,c]<>nil then
              DoBuildTOC(table.Cells[r,c].GetRVData, TOCList);
      end;
    end;
  end;
  {.........................................................}
begin
  Result := TRVList.Create;
  DoBuildTOC(rv.RVData, Result);
end;
{------------------------------------------------------------------------------}
procedure AddToc(rv: TCustomRichView; TOCList: TRVList;
  const TextStyleNos, RTFCodeTextStyleNos, ParaStyleNos: array of Integer;
  FirstItemToIncrement, Increment: Integer);
var i, Plus: Integer;
    TOCItem: TTOCItem;
    BookmarkName: TRVUnicodeString;
begin
  Plus := 0;
  for i := 0 to TOCList.Count-1 do begin
    if i=FirstItemToIncrement then
      Plus := Increment;
    TOCItem := TTOCItem(TOCList.Items[i]);
    rv.AddNLW(TOCItem.Text, TextStyleNos[TOCItem.Level], ParaStyleNos[TOCItem.Level]);
    rv.AddTab(TextStyleNos[TOCItem.Level], -1);
    if TOCItem.ID<0 then
      BookmarkName := '_toc'
    else
      BookmarkName := '_toc'+ RVIntToStrW(TOCItem.ID);
    rv.AddNLW('{\field{\*\fldinst PAGEREF '+BookmarkName+' \\h }{\fldrslt '+
      RVIntToStrW(TOCItem.PageNo+Plus)+'}}',
      RTFCodeTextStyleNos[TOCItem.Level], -1);
  end;
end;
{------------------------------------------------------------------------------}
// Returns index of the first TOC item after the TOC
function AddTOCTitleToTOC(rv: TCustomRichView; TitleParaNo, TOCPageNo, Depth: Integer;
  const Title: TRVUnicodeString; TOCList: TRVList): Integer;
var Level, i: Integer;
    TOCItem: TTOCItem;
begin
  Level := rv.Style.ParaStyles[TitleParaNo].OutlineLevel;
  Result := TOCList.Count;
  for i := 0 to TOCList.Count-1 do
    if TTOCItem(TOCList.Items[i]).PageNo>=TOCPageNo then begin
      Result := i;
      break;
      end;
  if (Level>0) and (Level<=Depth) then begin
    TOCItem := TTOCItem.Create;
    TOCItem.Text := Title;
    TOCItem.Level := Level-1;
    TOCItem.PageNo := TOCPageNo;
    TOCItem.ID := -1;
    TOCList.Insert(Result, TOCItem);
    inc(Result);
  end;
end;
{------------------------------------------------------------------------------}
function PrepareTOC(rv: TCustomRichView; RVPrint: TCustomRVPrint;
  ItemNo, TitleParaNo, Depth: Integer;
  const Title: TRVUnicodeString; var TOCList: TRVList;
  var TOCPageNo, TOCItemIndexAfterTOC: Integer): Boolean;
begin
  Result := (ItemNo=0) or (ItemNo=rv.ItemCount) or
    rv.PageBreaksBeforeItems[ItemNo];
  if not Result then
    exit;
  if ItemNo=rv.ItemCount then
    TOCPageNo := RVPrint.PagesCount+1
  else
    TOCPageNo := RVPrint.GetPageNo(rv.RVData, ItemNo, rv.GetOffsBeforeItem(ItemNo));
  TOCList := BuildTOCStructure(rv, RVPrint, Depth);
  if TOCList.Count=0 then begin
    TOCList.Free;
    Result := False;
    exit;
  end;
  TOCItemIndexAfterTOC := AddTOCTitleToTOC(rv, TitleParaNo, TOCPageNo, Depth,
    Title, TOCList);
end;
{------------------------------------------------------------------------------}
procedure PrepareRTFCodeTextStyleNos(const TextStyleNos: array of Integer;
  var RTFCodeTextStyleNos: array of Integer; rvs: TRVStyle);
var i, idx: Integer;
    TextStyle: TFontInfo;
begin
  TextStyle := TFontInfo.Create(nil);
  try
    for i := 0 to High(TextStyleNos) do
    begin
      TextStyle.Assign(rvs.TextStyles[TextStyleNos[i]]);
      TextStyle.Options := TextStyle.Options+[rvteoRTFCode];
      idx := rvs.FindTextStyle(TextStyle);
      RTFCodeTextStyleNos[i] := idx;
    end;
  finally
    TextStyle.Free;
  end;
end;

{------------------------------------------------------------------------------}
function AddTableOfContents(rv: TCustomRichView; RVPrint: TRVPrint;
  Depth: Integer; const TextStyleNos, ParaStyleNos: array of Integer;
  ItemNo: Integer; const Title: TRVUnicodeString;
  TitleStyleNo, TitleParaNo: Integer): Boolean;
var TOCList: TRVList;
    TOCPageCount: Integer;
    rv2: TCustomRichView;
    Stream: TRVMemoryStream;
    TOCPageNo, TOCItemIndexAfterTOC: Integer;
    RTFCodeTextStyleNos: array of Integer;
begin
  Result := PrepareTOC(rv, RVPrint, ItemNo, TitleParaNo, Depth, Title, TOCList,
    TOCPageNo, TOCItemIndexAfterTOC);
  if not Result then
    exit;
  SetLength(RTFCodeTextStyleNos, Length(TextStyleNos));
  PrepareRTFCodeTextStyleNos(TextStyleNos, RTFCodeTextStyleNos, rv.Style);
  if ItemNo<rv.ItemCount then
  begin
    rv2 := TRichView.Create(nil);
    try
      // building TOC in rv2, calculating TOCPageCount
      rv2.Visible := False;
      rv2.Parent := rv.Parent;
      rv2.Style := rv.Style;
      rv2.AddNLW(Title, TitleStyleNo, TitleParaNo);
      rv2.SetCheckpointInfo(rv2.ItemCount-1, RVEMPTYTAG, '_toc', False);
      AddTOC(rv2, TOCList, TextStyleNos, RTFCodeTextStyleNos, ParaStyleNos, TOCItemIndexAfterTOC, 1);
      RVPrint.AssignSource(rv2);
      RVPrint.FormatPages(rvdoAll);
      TOCPageCount := RVPrint.PagesCount;
      RVPrint.AssignSource(rv);
      // building TOC again, adding TOCPageCount+1 to page numbers
      rv2.Clear;
      rv2.AddNLW(Title, TitleStyleNo, TitleParaNo);
      rv2.SetCheckpointInfo(rv2.ItemCount-1, RVEMPTYTAG, '_toc', False);
      AddTOC(rv2, TOCList, TextStyleNos, RTFCodeTextStyleNos, ParaStyleNos, TOCItemIndexAfterTOC, TOCPageCount);
      // copying this TOC from rv2 to the rv, adding page break(s)
      Stream := TRVMemoryStream.Create;
      try
        if ItemNo>0 then
          rv2.PageBreaksBeforeItems[0] := True;
        rv.PageBreaksBeforeItems[ItemNo] := True;
        rv2.SaveRVFToStream(Stream, False);
        Stream.Position := 0;
        rv.InsertRVFFromStream(Stream, ItemNo);
      finally
        Stream.Free;
      end;
    finally
      rv2.Free;
    end;
  end else begin
    rv.AddNLW(Title, TitleStyleNo, TitleParaNo);
    rv.SetCheckpointInfo(rv.ItemCount-1, RVEMPTYTAG, '_toc', False);
    rv.PageBreaksBeforeItems[rv.ItemCount-1] := True;
    AddTOC(rv, TOCList, TextStyleNos, RTFCodeTextStyleNos, ParaStyleNos, 0, 0);
  end;
  RVPrint.Clear;
  TOCList.Free;
end;
{------------------------------------------------------------------------------}
function AddTableOfContents(RVReportHelper: TRVReportHelper; Canvas: TCanvas;
  PageWidth, FirstPageHeight, PageHeight: Integer;
  Depth: Integer; const TextStyleNos, ParaStyleNos: array of Integer;
  ItemNo: Integer; const Title: TRVUnicodeString; TitleStyleNo, TitleParaNo: Integer): Boolean;
var TOCList: TRVList;
    TOCPageCount: Integer;
    RVReportHelper2: TRVReportHelper;
    Stream: TRVMemoryStream;
    TOCPageNo, TOCItemIndexAfterTOC, LPageHeight: Integer;
    RTFCodeTextStyleNos: array of Integer;    
begin
  Result := PrepareTOC(RVReportHelper.RichView, RVReportHelper, ItemNo, TitleParaNo, Depth,
    Title, TOCList, TOCPageNo, TOCItemIndexAfterTOC);
  if not Result then
    exit;
  SetLength(RTFCodeTextStyleNos, Length(TextStyleNos));
  PrepareRTFCodeTextStyleNos(TextStyleNos, RTFCodeTextStyleNos, RVReportHelper.RichView.Style);
  if ItemNo<RVReportHelper.RichView.ItemCount then begin
    RVReportHelper2 := TRVReportHelper.Create(nil);
    try
      // building TOC in RVReportHelper2, calculating TOCPageCount
      RVReportHelper2.RichView.Style := RVReportHelper.RichView.Style;
      RVReportHelper2.RichView.LeftMargin   := RVReportHelper.RichView.LeftMargin;
      RVReportHelper2.RichView.TopMargin    := RVReportHelper.RichView.TopMargin;
      RVReportHelper2.RichView.RightMargin  := RVReportHelper.RichView.RightMargin;
      RVReportHelper2.RichView.BottomMargin := RVReportHelper.RichView.BottomMargin;      
      RVReportHelper2.RichView.AddNLW(Title, TitleStyleNo, TitleParaNo);
      RVReportHelper2.RichView.SetCheckpointInfo(RVReportHelper2.RichView.ItemCount-1, RVEMPTYTAG, '_toc', False);
      AddTOC(RVReportHelper2.RichView, TOCList, TextStyleNos, RTFCodeTextStyleNos,
        ParaStyleNos, TOCItemIndexAfterTOC, 1);
      RVReportHelper2.Init(Canvas, PageWidth);
      if ItemNo=0 then
        LPageHeight := FirstPageHeight
      else
        LPageHeight := PageHeight;
      while RVReportHelper2.FormatNextPage(LPageHeight) do
        LPageHeight := PageHeight;
      TOCPageCount := RVReportHelper2.PagesCount;
      // building TOC again, adding TOCPageCount+1 to page numbers
      RVReportHelper2.RichView.Clear;
      RVReportHelper2.RichView.AddNLW(Title, TitleStyleNo, TitleParaNo);
      RVReportHelper2.RichView.SetCheckpointInfo(RVReportHelper2.RichView.ItemCount-1, RVEMPTYTAG, '_toc', False);
      AddTOC(RVReportHelper2.RichView, TOCList, TextStyleNos, RTFCodeTextStyleNos,
        ParaStyleNos, TOCItemIndexAfterTOC, TOCPageCount);
      // copying this TOC from RVReportHelper2 to the RVReportHelper, adding page break(s)
      Stream := TRVMemoryStream.Create;
      try
        if ItemNo>0 then
          RVReportHelper2.RichView.PageBreaksBeforeItems[0] := True;
        RVReportHelper.RichView.PageBreaksBeforeItems[ItemNo] := True;
        RVReportHelper2.RichView.SaveRVFToStream(Stream, False);
        Stream.Position := 0;
        RVReportHelper.RichView.InsertRVFFromStream(Stream, ItemNo);
      finally
        Stream.Free;
      end;
    finally
      RVReportHelper2.Free;
    end;
  end else begin
    RVReportHelper.RichView.AddNLW(Title, TitleStyleNo, TitleParaNo);
    RVReportHelper.RichView.SetCheckpointInfo(RVReportHelper.RichView.ItemCount-1, RVEMPTYTAG, '_toc', False);    
    RVReportHelper.RichView.PageBreaksBeforeItems[RVReportHelper.RichView.ItemCount-1] := True;
    AddTOC(RVReportHelper.RichView, TOCList, TextStyleNos, RTFCodeTextStyleNos,
      ParaStyleNos, 0, 0);
  end;
  TOCList.Free;
end;
{------------------------------------------------------------------------------}
procedure GetStylesForOutlineLevel(rv: TCustomRichView; Level: Integer;
  var StyleNo, ParaNo: Integer);
  {.........................................................}
  function DoGetStyle(RVData: TCustomRVData): Boolean;
  var i, r, c: Integer;
      table: TRVTableItemInfo;
  begin
    Result := False;
    for i := 0 to RVData.ItemCount-1 do
    begin
      if (RVData.GetItemStyle(i)>=0) and
        (rv.Style.ParaStyles[RVData.GetItemPara(i)].OutlineLevel=Level) then
      begin
        Result := True;
        StyleNo := RVData.GetItemStyle(i);
        ParaNo := RVData.GetItemPara(i);
        exit;
      end;
      if RVData.GetItemStyle(i)=rvsTable then
      begin
        table := TRVTableItemInfo(RVData.GetItem(i));
        for r := 0 to table.RowCount-1 do
          for c := 0 to table.ColCount-1 do
            if table.Cells[r,c]<>nil then
            begin
              Result := DoGetStyle(table.Cells[r,c].GetRVData);
              if Result then
                exit;
            end;
      end;
    end;
  end;
  {.........................................................}
begin
  ParaNo := 0;
  StyleNo := 0;
  DoGetStyle(rv.RVData);
end;

end.
